#!/usr/bin/perl
###############################################################################
#                                                                             #
# Favorites                                                                   #
# v1.0                                                                        #
# (c) 2000 Anatoli Klassen                                                    #
#                                                                             #
###############################################################################

use strict;
use vars;

### Settings ##################################################################

my $pwd           = "fZZrAY8FkBKRM";      # crypted supervisor password - "edit" in this case
my $root_name;                            # root directory for all files, by default the directory
                                          # with the script, must finish with slash (for UNIX)
                                          # or backslash (for Windows)
my $links_name   = "favorites.txt";       # file to store links
my $su_view_name = "favorites_su.html";   # file with template for supervisor
my $edit_name    = "favorites_edit.html"; # file with template for edit
my $view_name    = "favorites.html";      # file with template for normal visitor

# HTML-codes for navigator labels
my $edit_title = "<img src=\"images/edit.gif\" alt=\"edit\" width=16 height=16 border=0>";

### Main Part #################################################################

# Global variables declaration

my %data;                 # data from HTTP request
my $supervisor;           # is in the supervisor mode
my $favorites_name;       # name of file with messages

my $file_name;            # name of file with the script
my $script_name;          # name of the script without directory
my $error;                # error message
my $edit;                 # in edit mode
my $item_value;           # value of item        for edit
my $caption_value;        # value of caption     for edit
my $link_value;           # value of link        for edit
my $description_value;    # value of description for edit

local *MSGFILE;

# Form name of files and directories

$file_name = __FILE__;
$script_name = $file_name;
$script_name =~ s|(.*/)||;
$script_name =~ s|(.*\\)||;

if (!defined($root_name)) {
  if ($file_name =~ m|\\|) {
    ($root_name) = ($file_name =~ m|(.*\\)|);
  }
  else {
    ($root_name) = ($file_name =~ m|(.*/)|);
  }
}

# Parse data

parse_data(\%data);

# Process

$supervisor = (crypt($data{"pwd"}, $pwd) eq $pwd) if ($data{"pwd"});
$pwd = $data{"pwd"} if ($supervisor);

$favorites_name = $root_name . $links_name;

if ($supervisor) {                    # for supervisor
  $error = do_supervisor();
}
else {                                # for normal visitor
  $error = do_visitor();
}
my @favorites = make_favorites($supervisor);

# Form the page

my $template_name;
my $output = "";

if ($supervisor) {
  if ($edit) {
    $template_name = $root_name . $edit_name;
  }
  else {
    $template_name = $root_name . $su_view_name;
  }
}
else {
  $template_name = $root_name . $view_name;
}

if (open (TEMPLATEFILE, "<$template_name")) {
  while (my $line = <TEMPLATEFILE>) {
    $output .= "$line";
  }
  close (TEMPLATEFILE);
}

if (!$output) {
  $output = "Internal server error.";
}

$output =~ s|<#main>|@favorites|gi;
$output =~ s/<#scriptname>/$script_name/gi;
$output =~ s/<#filename>/$file_name/gi;
$output =~ s/<#error>/$error/gi;
if ($supervisor) {
  $output =~ s/<#pwd>/$pwd/gi;
  if ($edit) {
    $output =~ s/<#item>/$item_value/gi;
    $output =~ s/<#caption>/$caption_value/gi;
    $output =~ s/<#link>/$link_value/gi;
    $output =~ s/<#desciption>/$description_value/gi;

    my $folder = get_folder();
    $folder = "/" if (!$folder);
    $folder = decode_string($folder);
    $output =~ s/<#folder>/$folder/gi;
  }
}
my $folder = get_folder();
$folder = "/" if (!$folder);
$output =~ s/<#folder>/$folder/gi;

# Output HTML

print "Content-type: text/html\nPragma: no-cache\n\n";
print $output;

# End of program

### Specific Subroutines ######################################################

sub get_folder
{
  my $name = shift @_;
  $name = "folder" if (!$name);

  my $folder = $data{"$name"};

  # force folder to begin with slash and to end without slash
  $folder = "" if (!defined($folder));
  $folder = substr($folder, 0, -1) if ($folder =~ m|/$|);
  $folder = "/".$folder if (!($folder =~ m|^/|) && ($folder ne ""));

  return $folder;
}

sub check_existing
{
  my $folder  = shift @_;
  my $caption = shift @_;
  return undef if (!defined($folder) || !defined($caption));

  my $folder_exists  = 0;
  my $caption_exists = 0;
  if (open (MSGFILE, "<$favorites_name")) {
    lock_file(\*MSGFILE);
    while (my $input = <MSGFILE>) {
      chomp $input;
      $folder_exists  ||= ($input =~ m|^$folder\"\"|);
      $caption_exists ||= ($input =~ m|^$folder/$caption\"|);
    }
    unlock_file(\*MSGFILE);
    close (MSGFILE);
  }
  $folder_exists = 1 if ($folder eq "");

  return ($folder_exists, $caption_exists)
}

sub make_favorites
{
  my $supervisor = shift @_;
  $supervisor = 0 if (!defined($supervisor));

  my $folder = get_folder();
  my @folders;

  # verify the given info
  my ($folder_exists, $caption_exists) = check_existing($folder, "");
  while (!(($folder_exists) = check_existing($folder, ""))) {
    ($folder) = ($folder =~ m|^(.*)/|);
  }

  if (open (MSGFILE, "<$favorites_name")) {
    lock_file(\*MSGFILE);
    while (my $input = <MSGFILE>) {
      chomp $input;
      push @folders, $input;
    }
    unlock_file(\*MSGFILE);
    close (MSGFILE);
  }
  else {
    return "<p>Internal server error.</p>";
  }

  my @tree;                                      # folders tree
  my @subfolders = split("/", $folder);
  unshift @subfolders, "/" if ($#subfolders < 0);

  for (my $i = 0; $i <= $#subfolders; $i++) {
    my $subfolder = "";
    for (my $j = 1; $j <= $i; $j++) {
      $subfolder .= "/".$subfolders[$j];
    }

    my @subtree;
    for (my $j = 0; $j <= $#folders; $j++) {
      my ($a_item, $a_link, $a_description) = ($folders[$j] =~ m|^$subfolder/([^/"]*)"([^"]*)"([^"]*)|);
      push @subtree, {
        "item" => $a_item,
        "link" => $a_link,
        "description" => $a_description
      } if ($a_item);
    }

    @{$tree[$i]} = sort compare_items @subtree;
  }

  shift @subfolders; # first element is always / or empty

  my $pwd_add;
  if($supervisor) {
    $pwd_add = "&pwd=$pwd";
  }

  return "<ul>\n<li class=\"opened\"><a href=\"$script_name?$pwd_add\" target=_self>[favorites]</a>\n",
         print_subtree(0, \@subfolders, \@tree, $supervisor),
         "</li>\n</ul>\n";
}

sub compare_items
{
  my %first  = %{$a};
  my %second = %{$b};

  if (($first{'link'} eq "") xor ($second{'link'} eq "")) {
    return ($first{'link'} eq "") ? -1 : 1;
  }
  else {
    return $first{'item'} cmp $second{'item'};
  }
}

sub print_subtree
{
  my $level      = shift @_;
  my @names      = @{shift @_};
  my @trees      = @{shift @_};
  my $supervisor = shift @_;

  $supervisor = 0 if (!defined($supervisor));
  my $name = $names[$level];
  my @tree = @{shift @trees};

  my @result;
  push @result, "<ul>\n";

  my $full_name = "%2F";
  for (my $i = 0; $i < $level; $i++) {
    $full_name .= $names[$i]."%2F";
  }

  while (my $node_ref = shift @tree) {
    my %node = %{$node_ref};
    my $edit_link;
    my $pwd_add;

    if($supervisor) {
      $pwd_add   = "&pwd=$pwd";
      $edit_link = " <a href=\"$script_name?pwd=$pwd&item=$full_name$node{'item'}\" target=_self>$edit_title</a>";
    }

    my $item_text        = decode_string($node{'item'});
    my $link_text        = decode_string($node{'link'});
    my $description_text = decode_string($node{'description'});

    if ($node{'link'}) {
      push @result,
        "<li class=\"link\" title=\"$description_text\"><a href=\"$link_text\">$item_text</a>$edit_link</li>\n";
    }
    else {
      my $item_class;
      if ($name eq $node{'item'}) {
        $item_class = "opened";
      }
      else {
        $item_class = "closed";
      }
      push @result,
        "<li class=\"$item_class\" title=\"$description_text\"><a href=\"$script_name?folder=$full_name$node{'item'}&$pwd_add\" target=_self>$item_text</a>$edit_link";

      push @result, print_subtree($level+1, \@names, \@trees, $supervisor) if ($name eq $node{'item'});
      push @result, "</li>\n";
    }
  }

  push @result, "</ul>\n";
  return @result;
}

sub do_visitor
{
  return undef;
}

sub do_supervisor
{
  if ($data{"item"}) {
    return undef if ($data{"cancel"});

    my $item = $data{"item"};
    my ($folder, $old_caption) = ($item =~ m|(.*)/(.*)|);

    $data{"folder"} = $folder;
    $folder = get_folder();

    # verify the given info
    my ($folder_exists, $item_exists) = check_existing($folder, $old_caption);
    if (!$item_exists || !$folder_exists) {
      return "The caption \"".decode_string($old_caption)."\" not found in folder \""
        .decode_string($folder)."\".";
    }

    my $caption;
    my $link;
    my $description;

    if (open (MSGFILE, "<$favorites_name")) {
      lock_file(\*MSGFILE);
      while (my $input = <MSGFILE>) {
        chomp $input;
        ($caption, $link, $description) = ($input =~ m|^$folder/($old_caption)\"([^"]*)"(.*)|);
        last if ($caption);
      }
      unlock_file(\*MSGFILE);
      close (MSGFILE);
    }

    if (!$data{"change"} && !$data{"delete"}) {
      $edit              = 1;
      $item_value        = $item;
      $caption_value     = decode_string($caption);
      $link_value        = decode_string($link);
      $description_value = decode_string($description);
    }
    elsif ($data{"delete"}) {
      my @items;
      if (open (MSGFILE, "<$favorites_name")) {
        lock_file(\*MSGFILE);
        while (my $input = <MSGFILE>) {
          push @items, $input if (!($input =~ m|^$item["/]|));
        }
        unlock_file(\*MSGFILE);
        close (MSGFILE);

        if (open (MSGFILE, ">$favorites_name")) {
          lock_file(\*MSGFILE);
          while (my $output = shift @items) {
            print MSGFILE $output;
          }
          unlock_file(\*MSGFILE);
          close (MSGFILE);
        }
      }
    }
    elsif ($data{"change"}) {
      my $new_folder      = encode_string2(get_folder("newfolder"));
      my $new_caption     = encode_string(secure_string($data{"caption"}));
      my $new_link        = encode_string(secure_string($data{"link"}));
      my $new_description = encode_string(secure_string($data{"description"}));

      # verify the given info
      my ($folder_exists, $item_exists) = check_existing($new_folder, $new_caption);
      if (!$folder_exists) {
        return "The folder \"".decode_string($new_folder)."\" not found.";
      }
      elsif ($item_exists && ($new_folder ne $folder)) {
        return "The caption \"".decode_string($new_caption)
          ."\" already exists in folder \"".decode_string($new_folder)."\".";
      }

      my @items;
      if (open (MSGFILE, "<$favorites_name")) {
        lock_file(\*MSGFILE);
        while (my $input = <MSGFILE>) {
          if ($input =~ m|^$item"|) {
            push @items, "$new_folder/$new_caption\"$new_link\"$new_description\n";
          }
          elsif (my ($subitem, $sublink, $subdescr) = $input =~ m|^$item/([^"]*)"([^"]*)"(.*)|) {
            push @items, "$new_folder/$new_caption/$subitem\"$sublink\"$subdescr\n";
          }
          else {
            push @items, $input;
          }
        }
        unlock_file(\*MSGFILE);
        close (MSGFILE);

        if (open (MSGFILE, ">$favorites_name")) {
          lock_file(\*MSGFILE);
          while (my $output = shift @items) {
            print MSGFILE $output;
          }
          unlock_file(\*MSGFILE);
          close (MSGFILE);
        }
      }
    }
  }

  elsif ($data{"link"} && $data{"create_link"} || $data{"caption"} && $data{"create_folder"}) {
    my $folder      = get_folder();
    my $caption     = encode_string(secure_string($data{"caption"}));
    my $link        = $data{"link"};
    my $description = encode_string(secure_string($data{"description"}, 1));
    my $output      = "";

    if (!$caption) {
      $caption = $data{"link"};
      $caption =~ s|^http://||i;
      $caption = encode_string(secure_string($caption));
    }

    if ($link) {
      $link = "http://" . $link if (!($link =~ m|^http://|));
      $link = encode_string(secure_string($link));
    }

    # verify the given info
    my ($folder_exists, $caption_exists) = check_existing($folder, $caption);

    if ($caption_exists) {
      return "The caption \"".decode_string($caption)."\" already exists in folder \""
        .decode_string($folder)."\".";
    }
    elsif (!$folder_exists) {
      return "The folder \"".decode_string($folder)."\" not found.";
    }
    else {
      if ($data{"link"} && $data{"create_link"}) {
        $output = "$folder/$caption\"$link\"$description\n";
      }
      elsif ($data{"caption"} && $data{"create_folder"}) {
        $output = "$folder/$caption\"\"$description\n";
      }

      if (open (MSGFILE, ">>$favorites_name")) {         # save the info
        lock_file(\*MSGFILE);
        print MSGFILE $output;
        unlock_file(\*MSGFILE);
        close (MSGFILE);
      }
      else {
        return "Can not open file \"$favorites_name\" to write.";
      }
    }
  }

  elsif ($data{"create_link"} && !$data{"link"}) {
    return "The link is not cpecified."
  }

  elsif ($data{"create_folder"} && $data{"caption"}) {
    return "The new folder name is not cpecified."
  }

  return undef;
}

### General Subroutines #######################################################

# Parse input data.
#
# Input:  one reference to hash for result
# Output: none

sub parse_data
{
  my $input;
  my $name;
  my $value;
  my @pairs;
  my ($result) = shift @_;
  return if (!$result);

  if ($ENV{'REQUEST_METHOD'} eq "GET")
  {
    $input = $ENV{'QUERY_STRING'};
  }
  else {
    read(STDIN, $input, $ENV{'CONTENT_LENGTH'});
  }

  @pairs = split /&/, $input;

  foreach my $pair (@pairs) {
    ($name, $value) = split /=/, $pair;
    $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
    $value =~ s/~!/ ~!/g;
    $value =~ s/\+/ /g;
    %{$result} -> {$name} = $value;
  }
}

# Lock file to avoid shared access. Wait until it is unlock.
#
# Input:  one file handle
# Output: none

sub lock_file
{
  my $handle = shift;
  return if (!$handle);
  until (flock($handle, 2)) {
    sleep .20;
  }
}

# Unlock file.
#
# Input:  one file handle
# Output: none

sub unlock_file
{
  my $handle = shift;
  return if (!$handle);
  flock($handle, 8);
}

# Substitute in given string symbols " < > & with their HTML equivalents.
# If second parameter is true then symbols ASCII 13 will convert to <br>.
#
# Input:  the string
#         is new lines allowed in the string (optional)
# Output: string without the symbols

sub secure_string
{
  my $string = shift @_;
  return $string if (!$string);
  my $new_lines = shift @_;

  $string =~ s/&/&amp\;/g;
  $string =~ s/"/&quot\;/g;
  $string =~ s/</&lt\;/g;
  $string =~ s/>/&gt\;/g;
  $string =~ s/\015/<br>/g if ($new_lines);
  $string =~ s/[\000-\037]/ /g;

  return $string;
}

# Substitute in given string special symbols with their hex codes.
#
# Input:  the string
# Output: string without the symbols

sub encode_string
{
  my $string = shift @_;
  return $string if (!$string);

  $string =~ s|([^A-Za-z0-9\-:.,])|"_".sprintf("%02X", ord($1))|eg;

  return $string;
}

# Substitute in given string special symbols except / with their hex codes.
#
# Input:  the string
# Output: string without the symbols

sub encode_string2
{
  my $string = shift @_;
  return $string if (!$string);

  $string =~ s|([^/A-Za-z0-9\-:.,])|"_".sprintf("%02X", ord($1))|eg;

  return $string;
}

# Substitute in given string symbol # and hex number after it with corresponding symbol.
#
# Input:  the string
# Output: converted string

sub decode_string
{
  my $string = shift @_;
  return $string if (!$string);

  $string =~ s|_([A-Za-z0-9][A-Za-z0-9])|pack("C", hex($1))|eg;

  return $string;
}

__END__

### EOF #######################################################################
